from collections.abc import Sequence

import numpy as np
import numpy.typing as npt

from laddu.amplitudes import (
    breit_wigner,
    common,
    kmatrix,
    phase_space,
    ylm,
    zlm,
)
from laddu.data import Dataset

class ParameterLike: ...

def AmplitudeOne() -> Expression: ...
def AmplitudeZero() -> Expression: ...
def amplitude_sum(
    amplitudes: Sequence[AmplitudeID | Expression]
    | Sequence[AmplitudeID]
    | Sequence[Expression],
) -> Expression: ...
def amplitude_product(
    amplitudes: Sequence[AmplitudeID | Expression]
    | Sequence[AmplitudeID]
    | Sequence[Expression],
) -> Expression: ...
def constant(value: float) -> ParameterLike: ...
def parameter(name: str) -> ParameterLike: ...

class AmplitudeID:
    def real(self) -> Expression: ...
    def imag(self) -> Expression: ...
    def conj(self) -> Expression: ...
    def norm_sqr(self) -> Expression: ...
    def __add__(self, other: AmplitudeID | Expression | int) -> Expression: ...
    def __radd__(self, other: AmplitudeID | Expression | int) -> Expression: ...
    def __sub__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __rsub__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __mul__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __rmul__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __truediv__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __rtruediv__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __neg__(self) -> Expression: ...

class Expression:
    def real(self) -> Expression: ...
    def imag(self) -> Expression: ...
    def conj(self) -> Expression: ...
    def norm_sqr(self) -> Expression: ...
    def __add__(self, other: AmplitudeID | Expression | int) -> Expression: ...
    def __radd__(self, other: AmplitudeID | Expression | int) -> Expression: ...
    def __sub__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __rsub__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __mul__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __rmul__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __truediv__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __rtruediv__(self, other: AmplitudeID | Expression) -> Expression: ...
    def __neg__(self) -> Expression: ...

class Amplitude: ...

class Manager:
    parameters: list[str]

    def __init__(self) -> None: ...
    def register(self, amplitude: Amplitude) -> AmplitudeID: ...
    def model(self, expression: Expression | AmplitudeID) -> Model: ...

class Model:
    parameters: list[str]

    def __init__(self) -> None: ...
    def load(self, dataset: Dataset) -> Evaluator: ...
    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...

class Evaluator:
    parameters: list[str]

    def activate(self, name: str | list[str]) -> None: ...
    def activate_all(self) -> None: ...
    def deactivate(self, name: str | list[str]) -> None: ...
    def deactivate_all(self) -> None: ...
    def isolate(self, name: str | list[str]) -> None: ...
    def evaluate(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        *,
        threads: int | None = None,
    ) -> npt.NDArray[np.complex128]: ...
    def evaluate_batch(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        indices: list[int] | npt.NDArray[np.integer],
        *,
        threads: int | None = None,
    ) -> npt.NDArray[np.complex128]: ...
    def evaluate_gradient(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        *,
        threads: int | None = None,
    ) -> npt.NDArray[np.complex128]: ...
    def evaluate_gradient_batch(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        indices: list[int] | npt.NDArray[np.integer],
        *,
        threads: int | None = None,
    ) -> npt.NDArray[np.complex128]: ...

def TestAmplitude(name: str, re: ParameterLike, im: ParameterLike) -> Amplitude: ...

__all__ = [
    'Amplitude',
    'AmplitudeID',
    'AmplitudeOne',
    'AmplitudeZero',
    'Evaluator',
    'Expression',
    'Manager',
    'Model',
    'ParameterLike',
    'TestAmplitude',
    'amplitude_sum',
    'breit_wigner',
    'common',
    'constant',
    'kmatrix',
    'parameter',
    'phase_space',
    'ylm',
    'zlm',
]



from laddu import amplitudes, convert, data, experimental, extensions, mpi, utils
from laddu.amplitudes import (
    AmplitudeOne,
    AmplitudeZero,
    Manager,
    Model,
    amplitude_product,
    amplitude_sum,
    constant,
    parameter,
)
from laddu.amplitudes.breit_wigner import BreitWigner
from laddu.amplitudes.common import ComplexScalar, PolarComplexScalar, Scalar
from laddu.amplitudes.phase_space import PhaseSpaceFactor
from laddu.amplitudes.ylm import Ylm
from laddu.amplitudes.zlm import PolPhase, Zlm
from laddu.convert import convert_from_amptools
from laddu.data import BinnedDataset, Dataset, Event, open, open_amptools
from laddu.extensions import (
    NLL,
    AutocorrelationTerminator,
    ControlFlow,
    EnsembleStatus,
    LikelihoodEvaluator,
    LikelihoodExpression,
    LikelihoodID,
    LikelihoodManager,
    LikelihoodOne,
    LikelihoodScalar,
    LikelihoodTerm,
    LikelihoodZero,
    MCMCObserver,
    MCMCSummary,
    MCMCTerminator,
    MinimizationObserver,
    MinimizationStatus,
    MinimizationSummary,
    MinimizationTerminator,
    StochasticNLL,
    Swarm,
    SwarmParticle,
    Walker,
    integrated_autocorrelation_times,
    likelihood_product,
    likelihood_sum,
)
from laddu.utils.variables import (
    Angles,
    CosTheta,
    Mandelstam,
    Mass,
    Phi,
    PolAngle,
    Polarization,
    PolMagnitude,
)
from laddu.utils.vectors import Vec3, Vec4

__version__: str

def available_parallelism() -> int: ...

__all__ = [
    'NLL',
    'AmplitudeOne',
    'AmplitudeZero',
    'Angles',
    'AutocorrelationTerminator',
    'BinnedDataset',
    'BreitWigner',
    'ComplexScalar',
    'ControlFlow',
    'CosTheta',
    'Dataset',
    'EnsembleStatus',
    'Event',
    'LikelihoodEvaluator',
    'LikelihoodExpression',
    'LikelihoodID',
    'LikelihoodManager',
    'LikelihoodOne',
    'LikelihoodScalar',
    'LikelihoodTerm',
    'LikelihoodZero',
    'MCMCObserver',
    'MCMCSummary',
    'MCMCTerminator',
    'Manager',
    'Mandelstam',
    'Mass',
    'MinimizationObserver',
    'MinimizationStatus',
    'MinimizationSummary',
    'MinimizationTerminator',
    'Model',
    'PhaseSpaceFactor',
    'Phi',
    'PolAngle',
    'PolMagnitude',
    'PolPhase',
    'PolarComplexScalar',
    'Polarization',
    'Scalar',
    'StochasticNLL',
    'Swarm',
    'SwarmParticle',
    'Vec3',
    'Vec4',
    'Walker',
    'Ylm',
    'Zlm',
    '__version__',
    'amplitude_product',
    'amplitude_sum',
    'amplitudes',
    'constant',
    'convert',
    'convert_from_amptools',
    'data',
    'experimental',
    'extensions',
    'integrated_autocorrelation_times',
    'likelihood_product',
    'likelihood_sum',
    'mpi',
    'open',
    'open_amptools',
    'parameter',
    'utils',
]



from typing import Literal

import numpy as np
import numpy.typing as npt

from laddu.data import Dataset, Event


class VariableExpression:
    def __and__(self, rhs: VariableExpression) -> VariableExpression: ...
    def __or__(self, rhs: VariableExpression) -> VariableExpression: ...
    def __invert__(self) -> VariableExpression: ...


class Mass:
    def __init__(self, constituents: list[int]) -> None: ...
    def value(self, event: Event) -> float: ...
    def value_on(self, dataset: Dataset) -> npt.NDArray[np.float64]: ...


class CosTheta:
    def __init__(
        self,
        beam: int,
        recoil: list[int],
        daughter: list[int],
        resonance: list[int],
        frame: Literal[
            'Helicity',
            'HX',
            'HEL',
            'GottfriedJackson',
            'Gottfried Jackson',
            'GJ',
            'Gottfried-Jackson',
        ] = 'Helicity',
    ) -> None: ...
    def value(self, event: Event) -> float: ...
    def value_on(self, dataset: Dataset) -> npt.NDArray[np.float64]: ...
    def __eq__(self, value: float) -> VariableExpression: ...
    def __lt__(self, value: float) -> VariableExpression: ...
    def __gt__(self, value: float) -> VariableExpression: ...
    def __le__(self, value: float) -> VariableExpression: ...
    def __ge__(self, value: float) -> VariableExpression: ...


class Phi:
    def __init__(
        self,
        beam: int,
        recoil: list[int],
        daughter: list[int],
        resonance: list[int],
        frame: Literal[
            'Helicity',
            'HX',
            'HEL',
            'GottfriedJackson',
            'Gottfried Jackson',
            'GJ',
            'Gottfried-Jackson',
        ] = 'Helicity',
    ) -> None: ...
    def value(self, event: Event) -> float: ...
    def value_on(self, dataset: Dataset) -> npt.NDArray[np.float64]: ...
    def __eq__(self, value: float) -> VariableExpression: ...
    def __lt__(self, value: float) -> VariableExpression: ...
    def __gt__(self, value: float) -> VariableExpression: ...
    def __le__(self, value: float) -> VariableExpression: ...
    def __ge__(self, value: float) -> VariableExpression: ...


class Angles:
    costheta: CosTheta
    phi: Phi

    def __init__(
        self,
        beam: int,
        recoil: list[int],
        daughter: list[int],
        resonance: list[int],
        frame: Literal[
            'Helicity',
            'HX',
            'HEL',
            'GottfriedJackson',
            'Gottfried Jackson',
            'GJ',
            'Gottfried-Jackson',
        ] = 'Helicity',
    ) -> None: ...


class PolAngle:
    def __init__(
        self,
        beam: int,
        recoil: list[int],
        beam_polarization: int,
    ) -> None: ...
    def value(self, event: Event) -> float: ...
    def value_on(self, dataset: Dataset) -> npt.NDArray[np.float64]: ...
    def __eq__(self, value: float) -> VariableExpression: ...
    def __lt__(self, value: float) -> VariableExpression: ...
    def __gt__(self, value: float) -> VariableExpression: ...
    def __le__(self, value: float) -> VariableExpression: ...
    def __ge__(self, value: float) -> VariableExpression: ...


class PolMagnitude:
    def __init__(
        self,
        beam_polarization: int,
    ) -> None: ...
    def value(self, event: Event) -> float: ...
    def value_on(self, dataset: Dataset) -> npt.NDArray[np.float64]: ...
    def __eq__(self, value: float) -> VariableExpression: ...
    def __lt__(self, value: float) -> VariableExpression: ...
    def __gt__(self, value: float) -> VariableExpression: ...
    def __le__(self, value: float) -> VariableExpression: ...
    def __ge__(self, value: float) -> VariableExpression: ...


class Polarization:
    pol_magnitude: PolMagnitude
    pol_angle: PolAngle

    def __init__(self, beam: int, recoil: list[int], beam_polarization: int) -> None: ...


class Mandelstam:
    def __init__(
        self,
        p1: list[int],
        p2: list[int],
        p3: list[int],
        p4: list[int],
        channel: Literal['s', 't', 'u'],
    ) -> None: ...
    def value(self, event: Event) -> float: ...
    def value_on(self, dataset: Dataset) -> npt.NDArray[np.float64]: ...
    def __eq__(self, value: float) -> VariableExpression: ...
    def __lt__(self, value: float) -> VariableExpression: ...
    def __gt__(self, value: float) -> VariableExpression: ...
    def __le__(self, value: float) -> VariableExpression: ...
    def __ge__(self, value: float) -> VariableExpression: ...



from typing import Sequence

import numpy as np
import numpy.typing as npt

__all__ = ['Vec3', 'Vec4']

class Vec3:
    """A 3-momentum vector formed from Cartesian components.

    Parameters
    ----------
    px, py, pz : float
        The Cartesian components of the 3-vector

    Attributes
    ----------
    mag : float
        The magnitude of the 3-vector
    mag2 : float
        The squared magnitude of the 3-vector
    """

    mag: float
    """
    The magnitude of the 3-vector

    .. math:: |\vec{p}| = \\sqrt{p_x^2 + p_y^2 + p_z^2}
    """
    mag2: float
    costheta: float
    theta: float
    phi: float
    unit: Vec3
    x: float
    y: float
    z: float
    px: float
    py: float
    pz: float

    def __init__(self, px: float, py: float, pz: float) -> None: ...
    def __add__(self, other: Vec3 | int) -> Vec3: ...
    def __radd__(self, other: Vec3 | int) -> Vec3: ...
    def __sub__(self, other: Vec3 | int) -> Vec3: ...
    def __rsub__(self, other: Vec3 | int) -> Vec3: ...
    def __mul__(self, other: float) -> Vec3: ...
    def __rmul__(self, other: float) -> Vec3: ...
    def __neg__(self) -> Vec3: ...
    def dot(self, other: Vec3) -> float:
        """Calculate the dot product of two vectors.

        Parameters
        ----------
        other : Vec3
            A vector input with which the dot product is taken

        Returns
        -------
        float
            The dot product of this vector and `other`
        """

    def cross(self, other: Vec3) -> Vec3:
        """
        Calculate the cross product of two vectors.

        Parameters
        ----------
        other : Vec3
            A vector input with which the cross product is taken

        Returns
        -------
        Vec3
            The cross product of this vector and `other`
        """

    def to_numpy(self) -> npt.NDArray[np.float64]: ...
    @staticmethod
    def from_array(array: Sequence) -> Vec3: ...
    def with_mass(self, mass: float) -> Vec4: ...
    def with_energy(self, mass: float) -> Vec4: ...

class Vec4:
    mag: float
    mag2: float
    vec3: Vec3
    t: float
    x: float
    y: float
    z: float
    e: float
    px: float
    py: float
    pz: float
    momentum: Vec3
    gamma: float
    beta: Vec3
    m: float
    m2: float

    def __init__(self, px: float, py: float, pz: float, e: float) -> None: ...
    def __add__(self, other: Vec4) -> Vec4: ...
    def __sub__(self, other: Vec4 | int) -> Vec4: ...
    def __rsub__(self, other: Vec4 | int) -> Vec4: ...
    def __mul__(self, other: float) -> Vec4: ...
    def __rmul__(self, other: float) -> Vec4: ...
    def __neg__(self) -> Vec4: ...
    def boost(self, beta: Vec3) -> Vec4: ...
    def to_numpy(self) -> npt.NDArray[np.float64]: ...
    @staticmethod
    def from_array(array: Sequence) -> Vec4: ...



from pathlib import Path
from typing import Any, overload

import numpy as np
import pandas as pd
import polars as pl
from numpy.typing import NDArray

from laddu.utils.variables import CosTheta, Mandelstam, Mass, Phi, PolAngle, PolMagnitude
from laddu.utils.variables import VariableExpression
from laddu.utils.vectors import Vec3, Vec4


def open_amptools(
    path: str | Path,
    tree: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
    boost_to_com: bool = True,
) -> Dataset: ...


class Event:
    p4s: list[Vec4]
    aux: list[Vec3]
    weight: float

    def __init__(
        self,
        p4s: list[Vec4],
        aux: list[Vec3],
        weight: float,
        *,
        rest_frame_indices: list[int] | None = None,
    ) -> None: ...
    def get_p4_sum(self, indices: list[int]) -> Vec4: ...
    def boost_to_rest_frame_of(self, indices: list[int]) -> Event: ...
    def evaluate(self, variable: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam) -> float: ...


class Dataset:
    events: list[Event]
    n_events: int
    n_events_weighted: float
    weights: NDArray[np.float64]

    def __init__(self, events: list[Event]) -> None: ...
    def __len__(self) -> int: ...
    def __add__(self, other: Dataset | int) -> Dataset: ...
    def __radd__(self, other: Dataset | int) -> Dataset: ...
    @overload
    def __getitem__(self, index: int) -> Event: ...
    @overload
    def __getitem__(
        self, index: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam
    ) -> NDArray[np.float64]: ...
    def __getitem__(
        self, index: int | Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam
    ) -> Event | NDArray[np.float64]: ...
    def bin_by(
        self,
        variable: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam,
        bins: int,
        range: tuple[float, float],
    ) -> BinnedDataset: ...
    def filter(self, expression: VariableExpression) -> Dataset: ...
    def bootstrap(self, seed: int) -> Dataset: ...
    def boost_to_rest_frame_of(self, indices: list[int]) -> Dataset: ...
    @staticmethod
    def from_dict(data: dict[str, Any], rest_frame_indices: list[int] | None = None) -> Dataset: ...
    @staticmethod
    def from_numpy(data: dict[str, NDArray[np.floating]], rest_frame_indices: list[int] | None = None) -> Dataset: ...
    @staticmethod
    def from_pandas(data: pd.DataFrame, rest_frame_indices: list[int] | None = None) -> Dataset: ...
    @staticmethod
    def from_polars(
        data: pl.DataFrame, rest_frame_indices: list[int] | None = None
    ) -> Dataset: ...
    def evaluate(
        self, variable: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam
    ) -> NDArray[np.float64]: ...


class BinnedDataset:
    n_bins: int
    range: tuple[float, float]
    edges: NDArray[np.float64]

    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> Dataset: ...


def open(path: str | Path, *, rest_frame_indices: list[int] | None = None) -> Dataset: ...



from types import TracebackType

def use_mpi(*, trigger: bool) -> None: ...
def finalize_mpi() -> None: ...
def using_mpi() -> bool: ...
def is_root() -> bool: ...
def get_rank() -> int: ...
def get_size() -> int: ...
def is_mpi_available() -> bool: ...

class MPI:
    """
    A context manager for MPI.

    Example
    -------
    .. code-block:: python

        import os
        from laddu.mpi import MPI

        def main():
            ...

        if __name__ == '__main__':
            with MPI(trigger=os.environ['MPI'] == '1'):
                main()
    """

    def __init__(self, *, trigger: bool = True) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None: ...

__all__ = [
    'MPI',
    'finalize_mpi',
    'get_rank',
    'get_size',
    'is_mpi_available',
    'is_root',
    'use_mpi',
    'using_mpi',
]



from abc import ABCMeta, abstractmethod
from collections.abc import Sequence
from enum import Enum
from typing import Literal

import numpy as np
import numpy.typing as npt

from laddu.amplitudes import Evaluator, Model
from laddu.data import Dataset

def likelihood_sum(
    likelihoods: Sequence[LikelihoodID | LikelihoodExpression]
    | Sequence[LikelihoodID]
    | Sequence[LikelihoodExpression],
) -> LikelihoodExpression: ...
def likelihood_product(
    likelihoods: Sequence[LikelihoodID | LikelihoodExpression]
    | Sequence[LikelihoodID]
    | Sequence[LikelihoodExpression],
) -> LikelihoodExpression: ...
def LikelihoodOne() -> LikelihoodExpression: ...
def LikelihoodZero() -> LikelihoodExpression: ...

class LikelihoodID:
    def __add__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __radd__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __mul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...
    def __rmul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...

class LikelihoodExpression:
    def __add__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __radd__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __mul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...
    def __rmul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...

class LikelihoodTerm: ...

class LikelihoodManager:
    def __init__(self) -> None: ...
    def register(
        self, likelihood_term: LikelihoodTerm, *, name: str | None = None
    ) -> LikelihoodID: ...
    def load(
        self, likelihood_expression: LikelihoodExpression
    ) -> LikelihoodEvaluator: ...

class MinimizationStatus:
    x: npt.NDArray[np.float64]
    fx: float
    message: str
    err: npt.NDArray[np.float64] | None
    n_f_evals: int
    n_g_evals: int
    cov: npt.NDArray[np.float64] | None
    hess: npt.NDArray[np.float64] | None
    converged: bool
    swarm: Swarm | None

class MinimizationSummary:
    bounds: list[tuple[float, float]] | None
    parameter_names: list[str] | None
    message: str
    x0: npt.NDArray[np.float64]
    x: npt.NDArray[np.float64]
    std: npt.NDArray[np.float64]
    fx: float
    cost_evals: int
    gradient_evals: int
    converged: bool
    covariance: npt.NDArray[np.float64]

    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...

class MCMCSummary:
    bounds: list[tuple[float, float]] | None
    parameter_names: list[str] | None
    message: str
    cost_evals: int
    gradient_evals: int
    converged: bool
    dimension: tuple[int, int, int]

    def get_chain(
        self, *, burn: int | None = None, thin: int | None = None
    ) -> npt.NDArray[np.float64]: ...
    def get_flat_chain(
        self, *, burn: int | None = None, thin: int | None = None
    ) -> npt.NDArray[np.float64]: ...
    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...

class EnsembleStatus:
    message: str
    n_f_evals: int
    n_g_evals: int
    walkers: list[Walker]
    dimension: tuple[int, int, int]

    def get_chain(
        self, *, burn: int | None = None, thin: int | None = None
    ) -> npt.NDArray[np.float64]: ...
    def get_flat_chain(
        self, *, burn: int | None = None, thin: int | None = None
    ) -> npt.NDArray[np.float64]: ...

class Swarm:
    particles: list[SwarmParticle]

class SwarmParticle:
    x: npt.NDArray[np.float64]
    fx: float
    x_best: npt.NDArray[np.float64]
    fx_best: float
    velocity: npt.NDArray[np.float64]

class Walker:
    dimension: tuple[int, int]

    def get_latest(self) -> tuple[npt.NDArray[np.float64], float]: ...

class ControlFlow(Enum):
    Continue = 0
    Break = 1

class MinimizationObserver(metaclass=ABCMeta):
    @abstractmethod
    def observe(self, step: int, status: MinimizationStatus) -> None: ...

class MinimizationTerminator(metaclass=ABCMeta):
    @abstractmethod
    def check_for_termination(
        self, step: int, status: MinimizationStatus
    ) -> ControlFlow: ...

class MCMCObserver(metaclass=ABCMeta):
    @abstractmethod
    def observe(self, step: int, status: EnsembleStatus) -> None: ...

class MCMCTerminator(metaclass=ABCMeta):
    @abstractmethod
    def check_for_termination(self, step: int, status: EnsembleStatus) -> ControlFlow: ...

class LikelihoodEvaluator:
    parameters: list[str]

    def evaluate(
        self,
        parameters: list[float] | npt.ArrayLike,
        threads: int | None = None,
    ) -> float: ...
    def evaluate_gradient(
        self,
        parameters: list[float] | npt.ArrayLike,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def minimize(
        self,
        p0: list[float] | npt.ArrayLike,
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['lbfgsb', 'nelder-mead', 'adam', 'pso'] = 'lbfgsb',
        settings: dict | None = None,
        observers: MinimizationObserver | Sequence[MinimizationObserver] | None = None,
        terminators: MinimizationTerminator
        | Sequence[MinimizationTerminator]
        | None = None,
        max_steps: int | None = None,
        debug: bool = False,
        threads: int = 0,
    ) -> MinimizationSummary: ...
    def mcmc(
        self,
        p0: list[list[float]] | npt.ArrayLike,
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['aies', 'ess'] = 'aies',
        settings: dict | None = None,
        observers: MCMCObserver | Sequence[MCMCObserver] | None = None,
        terminators: MCMCTerminator
        | AutocorrelationTerminator
        | Sequence[MCMCTerminator | AutocorrelationTerminator]
        | None = None,
        max_steps: int | None = None,
        debug: bool = False,
        threads: int = 0,
    ) -> MCMCSummary: ...

class StochasticNLL:
    nll: NLL

    def minimize(
        self,
        p0: list[float] | npt.ArrayLike,
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['lbfgsb', 'nelder-mead', 'adam', 'pso'] = 'lbfgsb',
        settings: dict | None = None,
        observers: MinimizationObserver | Sequence[MinimizationObserver] | None = None,
        terminators: MinimizationTerminator
        | Sequence[MinimizationTerminator]
        | None = None,
        max_steps: int | None = None,
        debug: bool = False,
        threads: int = 0,
    ) -> MinimizationSummary: ...
    def mcmc(
        self,
        p0: list[list[float]] | npt.ArrayLike,
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['aies', 'ess'] = 'aies',
        settings: dict | None = None,
        observers: MCMCObserver | Sequence[MCMCObserver] | None = None,
        terminators: MCMCTerminator
        | AutocorrelationTerminator
        | Sequence[MCMCTerminator | AutocorrelationTerminator]
        | None = None,
        max_steps: int | None = None,
        debug: bool = False,
        threads: int = 0,
    ) -> MCMCSummary: ...

class NLL:
    parameters: list[str]
    data: Dataset
    accmc: Dataset

    def __init__(
        self,
        model: Model,
        ds_data: Dataset,
        ds_accmc: Dataset,
    ) -> None: ...
    def to_term(self) -> LikelihoodTerm: ...
    def to_stochastic(
        self, batch_size: int, *, seed: int | None = None
    ) -> StochasticNLL: ...
    def activate(self, name: str | list[str]) -> None: ...
    def activate_all(self) -> None: ...
    def deactivate(self, name: str | list[str]) -> None: ...
    def deactivate_all(self) -> None: ...
    def isolate(self, name: str | list[str]) -> None: ...
    def evaluate(
        self,
        parameters: list[float] | npt.ArrayLike,
        threads: int | None = None,
    ) -> float: ...
    def evaluate_gradient(
        self,
        parameters: list[float] | npt.ArrayLike,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def project(
        self,
        parameters: list[float] | npt.ArrayLike,
        *,
        mc_evaluator: Evaluator | None = None,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def project_with(
        self,
        parameters: list[float] | npt.ArrayLike,
        name: str | list[str],
        *,
        mc_evaluator: Evaluator | None = None,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def minimize(
        self,
        p0: list[float] | npt.ArrayLike,
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['lbfgsb', 'nelder-mead', 'adam', 'pso'] = 'lbfgsb',
        settings: dict | None = None,
        observers: MinimizationObserver | Sequence[MinimizationObserver] | None = None,
        terminators: MinimizationTerminator
        | Sequence[MinimizationTerminator]
        | None = None,
        max_steps: int | None = None,
        debug: bool = False,
        threads: int = 0,
    ) -> MinimizationSummary: ...
    def mcmc(
        self,
        p0: list[list[float]] | npt.ArrayLike,
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['aies', 'ess'] = 'aies',
        settings: dict | None = None,
        observers: MCMCObserver | Sequence[MCMCObserver] | None = None,
        terminators: MCMCTerminator
        | AutocorrelationTerminator
        | Sequence[MCMCTerminator | AutocorrelationTerminator]
        | None = None,
        max_steps: int | None = None,
        debug: bool = False,
        threads: int = 0,
    ) -> MCMCSummary: ...

def LikelihoodScalar(name: str) -> LikelihoodTerm: ...

class AutocorrelationTerminator:
    taus: npt.NDArray[np.float64]

    def __init__(
        self,
        *,
        n_check: int = 50,
        n_taus_threshold: int = 50,
        dtau_threshold: float = 0.01,
        discard: float = 0.5,
        terminate: bool = True,
        c: float = 7.0,
        verbose: bool = False,
    ) -> None: ...

def integrated_autocorrelation_times(
    samples: npt.ArrayLike, *, c: float | None = None
) -> npt.NDArray[np.float64]: ...

__all__ = [
    'NLL',
    'AutocorrelationTerminator',
    'ControlFlow',
    'EnsembleStatus',
    'LikelihoodEvaluator',
    'LikelihoodExpression',
    'LikelihoodID',
    'LikelihoodManager',
    'LikelihoodOne',
    'LikelihoodScalar',
    'LikelihoodTerm',
    'LikelihoodZero',
    'MCMCObserver',
    'MCMCSummary',
    'MCMCTerminator',
    'MinimizationObserver',
    'MinimizationStatus',
    'MinimizationSummary',
    'MinimizationTerminator',
    'StochasticNLL',
    'Swarm',
    'SwarmParticle',
    'Walker',
    'integrated_autocorrelation_times',
    'likelihood_product',
    'likelihood_sum',
]



# pyright: reportUnknownMemberType=false, reportUnknownVariableType=false, reportMissingTypeStubs=false, reportAttributeAccessIssue=false, reportUnknownArgumentType=false
"""
Usage:
    amptools-to-laddu <input_file> <output_file> [--tree <treename>] [--pol-in-beam | --pol-angle <angle> --pol-magnitude <magnitude>] [-n <num-entries>]

Options:
    --tree <treename>            The tree name in the ROOT file [default: kin].
    --pol-in-beam                Use the beam's momentum for polarization (eps).
    --pol-angle <angle>          The polarization angle in degrees (only used if --pol-in-beam is not used)
    --pol-magnitude <magnitude>  The polarization magnitude (only used if --pol-in-beam is not used)
    -n <num-entries>             Truncate the file to the first n entries for testing.
"""  # noqa: D205, D400

from __future__ import annotations

from pathlib import Path

import numpy as np
import numpy.typing as npt
import polars as pl
import uproot
from docopt import docopt


def read_root_file(
    input_path: Path | str,
    tree_name: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle_rad: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    input_path = Path(input_path)
    tree = uproot.open(f'{input_path}:{tree_name}')  #
    E_beam: npt.NDArray[np.float32] = tree['E_Beam'].array(
        library='np', entry_stop=num_entries
    )
    Px_beam: npt.NDArray[np.float32] = tree['Px_Beam'].array(
        library='np', entry_stop=num_entries
    )
    Py_beam: npt.NDArray[np.float32] = tree['Py_Beam'].array(
        library='np', entry_stop=num_entries
    )
    Pz_beam: npt.NDArray[np.float32] = tree['Pz_Beam'].array(
        library='np', entry_stop=num_entries
    )
    weight = (
        tree['Weight'].array(library='np', entry_stop=num_entries)
        if 'Weight' in tree
        else np.ones_like(E_beam)
    )

    E_final: npt.NDArray[np.float32] = np.array(
        list(tree['E_FinalState'].array(library='np', entry_stop=num_entries))
    )
    Px_final: npt.NDArray[np.float32] = np.array(
        list(tree['Px_FinalState'].array(library='np', entry_stop=num_entries))
    )
    Py_final: npt.NDArray[np.float32] = np.array(
        list(tree['Py_FinalState'].array(library='np', entry_stop=num_entries))
    )
    Pz_final: npt.NDArray[np.float32] = np.array(
        list(tree['Pz_FinalState'].array(library='np', entry_stop=num_entries))
    )

    p4_beam = np.stack([Px_beam, Py_beam, Pz_beam, E_beam], axis=-1)
    p4_final = np.stack([Px_final, Py_final, Pz_final, E_final], axis=-1)

    if 'EPS' in tree:
        eps = tree['EPS'].array(library='np', entry_stop=num_entries)
        eps = eps[:, np.newaxis, :]
    if 'eps' in tree:
        eps = tree['eps'].array(library='np', entry_stop=num_entries)
        eps = eps[:, np.newaxis, :]
    elif pol_in_beam:
        eps = np.stack([Px_beam, Py_beam, Pz_beam], axis=-1)[:, np.newaxis]
        p4_beam[:, 0] = 0  # Set Px to 0
        p4_beam[:, 1] = 0  # Set Py to 0
        p4_beam[:, 2] = E_beam  # Set Pz = E for beam
    elif pol_angle_rad is not None and pol_magnitude is not None:
        eps_x = pol_magnitude * np.cos(pol_angle_rad) * np.ones_like(E_beam)
        eps_y = pol_magnitude * np.sin(pol_angle_rad) * np.ones_like(E_beam)
        eps_z = np.zeros_like(E_beam)
        eps = np.stack([eps_x, eps_y, eps_z], axis=-1)[:, np.newaxis]
    else:
        eps = np.zeros((len(E_beam), 1, 3), dtype=np.float32)  # Default to 0

    p4s = np.concatenate([p4_beam[:, np.newaxis, :], p4_final], axis=1)

    return p4s.astype(np.float32), eps.astype(np.float32), weight


def save_as_parquet(
    p4s: npt.NDArray[np.float32],
    eps: npt.NDArray[np.float32],
    weight: npt.NDArray[np.float32],
    output_path: Path | str,
) -> None:
    columns = {}
    n_particles = p4s.shape[1]
    for i in range(n_particles):
        columns[f'p4_{i}_Px'] = p4s[:, i, 0]
        columns[f'p4_{i}_Py'] = p4s[:, i, 1]
        columns[f'p4_{i}_Pz'] = p4s[:, i, 2]
        columns[f'p4_{i}_E'] = p4s[:, i, 3]

    n_eps = eps.shape[1]
    for i in range(n_eps):
        columns[f'aux_{i}_x'] = eps[:, i, 0]
        columns[f'aux_{i}_y'] = eps[:, i, 1]
        columns[f'aux_{i}_z'] = eps[:, i, 2]

    columns['weight'] = weight

    dataframe = pl.DataFrame(columns)
    dataframe.write_parquet(str(output_path))


def convert_from_amptools(
    input_path: Path,
    output_path: Path,
    tree_name: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
) -> None:
    p4s, eps, weight = read_root_file(
        input_path,
        tree_name,
        pol_in_beam=pol_in_beam,
        pol_angle_rad=pol_angle,
        pol_magnitude=pol_magnitude,
        num_entries=num_entries,
    )
    save_as_parquet(p4s, eps, weight, output_path)


def run() -> None:
    args = docopt(__doc__ if __doc__ else '')
    input_file = args['<input_file>']
    output_file = args['<output_file>']
    tree_name = args['--tree']
    pol_in_beam = args['--pol-in-beam']
    pol_angle = float(args['--pol-angle']) * np.pi / 180 if args['--pol-angle'] else None
    pol_magnitude = float(args['--pol-magnitude']) if args['--pol-magnitude'] else None
    num_entries = int(args['-n']) if args['-n'] else None

    convert_from_amptools(
        Path(input_file),
        Path(output_file),
        tree_name,
        pol_in_beam=pol_in_beam,
        pol_angle=pol_angle,
        pol_magnitude=pol_magnitude,
        num_entries=num_entries,
    )



from laddu.laddu import (
    Angles,
    CosTheta,
    Mandelstam,
    Mass,
    Phi,
    PolAngle,
    Polarization,
    PolMagnitude,
    VariableExpression,
)

__all__ = [
    'Angles',
    'CosTheta',
    'Mandelstam',
    'Mass',
    'Phi',
    'PolAngle',
    'PolMagnitude',
    'Polarization',
    'VariableExpression',
]



from laddu.amplitudes import Amplitude, ParameterLike

def Scalar(name: str, value: ParameterLike) -> Amplitude: ...
def ComplexScalar(name: str, re: ParameterLike, im: ParameterLike) -> Amplitude: ...
def PolarComplexScalar(
    name: str, r: ParameterLike, theta: ParameterLike
) -> Amplitude: ...



from __future__ import annotations

from typing import TYPE_CHECKING

from laddu.laddu import (
    finalize_mpi,
    get_rank,
    get_size,
    is_mpi_available,
    is_root,
    use_mpi,
    using_mpi,
)

if TYPE_CHECKING:
    from types import TracebackType


class MPI:
    r"""
    Context manager for using the MPI backend.

    Methods
    -------
    __init__(self, *, trigger: bool = True)
        Initialize the MPI context manager to use MPI if `trigger` is True.

    Examples
    --------
    .. code-block:: python

        import os
        from laddu.mpi import MPI

        def main():
            ...

        if __name__ == '__main__':
            with MPI(trigger=os.environ['MPI'] == '1'):
                main()

    This code will use MPI if and only if the environment variable "MPI" is set to "1", and it will properly finalize MPI when the context manager exits.

    """

    def __init__(self, *, trigger: bool = True) -> None:
        self._trigger = trigger

    def __enter__(self) -> None:
        use_mpi(trigger=self._trigger)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        finalize_mpi()


__all__ = [
    'MPI',
    'finalize_mpi',
    'get_rank',
    'get_size',
    'is_mpi_available',
    'is_root',
    'use_mpi',
    'using_mpi',
]



from __future__ import annotations

from abc import ABCMeta, abstractmethod

from laddu.laddu import (
    NLL,
    AutocorrelationTerminator,
    ControlFlow,
    EnsembleStatus,
    LikelihoodEvaluator,
    LikelihoodExpression,
    LikelihoodID,
    LikelihoodManager,
    LikelihoodOne,
    LikelihoodScalar,
    LikelihoodTerm,
    LikelihoodZero,
    MCMCSummary,
    MinimizationStatus,
    MinimizationSummary,
    StochasticNLL,
    Swarm,
    SwarmParticle,
    Walker,
    integrated_autocorrelation_times,
    likelihood_product,
    likelihood_sum,
)


class MinimizationObserver(metaclass=ABCMeta):
    @abstractmethod
    def observe(self, step: int, status: MinimizationStatus) -> None:
        pass


class MinimizationTerminator(metaclass=ABCMeta):
    @abstractmethod
    def check_for_termination(self, step: int, status: MinimizationStatus) -> ControlFlow:
        pass


class MCMCObserver(metaclass=ABCMeta):
    @abstractmethod
    def observe(self, step: int, status: EnsembleStatus) -> None:
        pass


class MCMCTerminator(metaclass=ABCMeta):
    @abstractmethod
    def check_for_termination(self, step: int, status: EnsembleStatus) -> ControlFlow:
        pass


__all__ = [
    'NLL',
    'AutocorrelationTerminator',
    'ControlFlow',
    'EnsembleStatus',
    'LikelihoodEvaluator',
    'LikelihoodExpression',
    'LikelihoodID',
    'LikelihoodManager',
    'LikelihoodOne',
    'LikelihoodScalar',
    'LikelihoodTerm',
    'LikelihoodZero',
    'MCMCObserver',
    'MCMCSummary',
    'MCMCTerminator',
    'MinimizationObserver',
    'MinimizationStatus',
    'MinimizationSummary',
    'MinimizationTerminator',
    'StochasticNLL',
    'Swarm',
    'SwarmParticle',
    'Walker',
    'integrated_autocorrelation_times',
    'likelihood_product',
    'likelihood_sum',
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np

from laddu.convert import read_root_file
from laddu.laddu import BinnedDataset, DatasetBase, Event, open
from laddu.utils.vectors import Vec3, Vec4

if TYPE_CHECKING:
    from pathlib import Path

    import pandas as pd
    import polars as pl
    from numpy.typing import NDArray


class Dataset(DatasetBase):
    P4_PREFIX: str = 'p4_'
    AUX_PREFIX: str = 'aux_'

    @staticmethod
    def from_dict(
        data: dict[str, Any], rest_frame_indices: list[int] | None = None
    ) -> Dataset:
        """
        Create a Dataset from a dict.

        Arguments
        ---------
        data: dict of str to Any
            dict of lists with keys matching the dataset format
        rest_frame_indices: list of int, optional
            If provided, the dataset will be boosted to the rest frame
            of the 4-momenta specified by the indices

        Returns
        -------
        Dataset
        """
        p4_count = len([key for key in data if key.startswith(Dataset.P4_PREFIX)]) // 4
        aux_count = len([key for key in data if key.startswith(Dataset.AUX_PREFIX)]) // 3
        events = []
        n_events = len(data[f'{Dataset.P4_PREFIX}0_Px'])
        weights = data.get('weight', np.ones(n_events))
        for i in range(n_events):
            p4s = [
                Vec4.from_array(
                    [
                        data[f'{Dataset.P4_PREFIX}{j}_Px'][i],
                        data[f'{Dataset.P4_PREFIX}{j}_Py'][i],
                        data[f'{Dataset.P4_PREFIX}{j}_Pz'][i],
                        data[f'{Dataset.P4_PREFIX}{j}_E'][i],
                    ]
                )
                for j in range(p4_count)
            ]
            aux = [
                Vec3.from_array(
                    [
                        data[f'{Dataset.AUX_PREFIX}{j}_x'][i],
                        data[f'{Dataset.AUX_PREFIX}{j}_y'][i],
                        data[f'{Dataset.AUX_PREFIX}{j}_z'][i],
                    ]
                )
                for j in range(aux_count)
            ]
            weight = weights[i]
            events.append(Event(p4s, aux, weight, rest_frame_indices=rest_frame_indices))
        return DatasetBase(events)

    @staticmethod
    def from_numpy(
        data: dict[str, NDArray[np.floating]], rest_frame_indices: list[int] | None = None
    ) -> Dataset:
        """
        Create a Dataset from a dict of numpy arrays.

        Arguments
        ---------
        data: dict of str to NDArray
            dict of arrays with keys matching the dataset format
        rest_frame_indices: list of int, optional
            If provided, the dataset will be boosted to the rest frame
            of the 4-momenta specified by the indices

        Returns
        -------
        Dataset
        """
        return Dataset.from_dict(data, rest_frame_indices=rest_frame_indices)

    @staticmethod
    def from_pandas(
        data: pd.DataFrame, rest_frame_indices: list[int] | None = None
    ) -> Dataset:
        """
        Create a Dataset from a pandas DataFrame.

        Arguments
        ---------
        data: pandas.DataFrame
            DataFrame with columns matching the dataset format
        rest_frame_indices: list of int, optional
            If provided, the dataset will be boosted to the rest frame
            of the 4-momenta specified by the indices

        Returns
        -------
        Dataset
        """
        return Dataset.from_dict(data.to_dict(), rest_frame_indices=rest_frame_indices)

    @staticmethod
    def from_polars(
        data: pl.DataFrame, rest_frame_indices: list[int] | None = None
    ) -> Dataset:
        """
        Create a Dataset from a polars DataFrame.

        Arguments
        ---------
        data: polars.DataFrame
            DataFrame with columns matching the dataset format
        rest_frame_indices: list of int, optional
            If provided, the dataset will be boosted to the rest frame
            of the 4-momenta specified by the indices

        Returns
        -------
        Dataset
        """
        return Dataset.from_dict(data.to_dict(), rest_frame_indices=rest_frame_indices)


def open_amptools(
    path: str | Path,
    tree: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
    boost_to_com: bool = True,
) -> Dataset:
    pol_angle_rad = pol_angle * np.pi / 180 if pol_angle else None
    p4s_list, eps_list, weight_list = read_root_file(
        path,
        tree,
        pol_in_beam=pol_in_beam,
        pol_angle_rad=pol_angle_rad,
        pol_magnitude=pol_magnitude,
        num_entries=num_entries,
    )
    n_particles = len(p4s_list[0])
    rest_frame_indices = list(range(1, n_particles)) if boost_to_com else None
    ds = Dataset(
        [
            Event(
                [Vec4.from_array(p4) for p4 in p4s],
                [Vec3.from_array(eps_vec) for eps_vec in eps],
                weight,
                rest_frame_indices=rest_frame_indices,
            )
            for p4s, eps, weight in zip(p4s_list, eps_list, weight_list)
        ]
    )
    return ds


__all__ = ['BinnedDataset', 'Dataset', 'Event', 'open', 'open_amptools']



from laddu.amplitudes import Amplitude, ParameterLike
from laddu.utils.variables import CosTheta, Mandelstam, Mass, Phi, PolAngle, PolMagnitude

def PiecewiseScalar(
    name: str,
    variable: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam,
    bins: int,
    range: tuple[float, float],
    values: list[ParameterLike],
) -> Amplitude: ...
def PiecewiseComplexScalar(
    name: str,
    variable: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam,
    bins: int,
    range: tuple[float, float],
    values: list[tuple[ParameterLike, ParameterLike]],
) -> Amplitude: ...
def PiecewisePolarComplexScalar(
    name: str,
    variable: Mass | CosTheta | Phi | PolAngle | PolMagnitude | Mandelstam,
    bins: int,
    range: tuple[float, float],
    values: list[tuple[ParameterLike, ParameterLike]],
) -> Amplitude: ...
